@@ -1,6 +1,6 @@
 /* GdkQuartzWindow.m
  *
- * Copyright (C) 2005 Imendio AB
+ * Copyright (C) 2005-2007 Imendio AB
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -18,7 +18,6 @@
  * Boston, MA 02111-1307, USA.
  */
 
-
 #import "GdkQuartzWindow.h"
 #include "gdkwindow-quartz.h"
 #include "gdkprivate-quartz.h"
@@ -69,6 +68,40 @@
   _gdk_quartz_events_update_focus_window (window, FALSE);
 }
 
+/* Used in combination with NSLeftMouseUp in sendEvent to keep track
+ * of when the window is being moved with the mouse.
+ */
+-(void)windowWillMove:(NSNotification *)aNotification
+{
+  if (leftDown)
+    inMove = YES;
+}
+
+-(void)sendEvent:(NSEvent *)event
+{
+  switch ([event type])
+    {
+    case NSLeftMouseDown:
+      leftDown = YES;
+      break;
+
+    case NSLeftMouseUp:
+      leftDown = NO;
+      inMove = NO;
+      break;
+
+    default:
+      break;
+    }
+
+  [super sendEvent:event];
+}
+
+-(BOOL)isInMove
+{
+  return inMove;
+}
+
 -(void)windowDidMove:(NSNotification *)aNotification
 {
   NSRect content_rect = [self contentRectForFrameRect:[self frame]];
@@ -102,8 +135,9 @@
   impl->width = content_rect.size.width;
   impl->height = content_rect.size.height;
 
-  /* Synthesize a configure event */
+  [[self contentView] setBounds:NSMakeRect (0, 0, impl->width, impl->height)];
 
+  /* Synthesize a configure event */
   event = gdk_event_new (GDK_CONFIGURE);
   event->configure.window = g_object_ref (window);
   event->configure.x = private->x;
@@ -112,13 +146,6 @@
   event->configure.height = impl->height;
 
   _gdk_event_queue_append (gdk_display_get_default (), event);
-
-  /* Update tracking rectangle */
-  [[self contentView] removeTrackingRect:impl->tracking_rect];
-  impl->tracking_rect = [impl->view addTrackingRect:NSMakeRect(0, 0, impl->width, impl->height) 
-			                      owner:impl->view
-			                   userData:nil
-			               assumeInside:NO];
 }
 
 -(id)initWithContentRect:(NSRect)contentRect styleMask:(unsigned int)styleMask backing:(NSBackingStoreType)backingType defer:(BOOL)flag
@@ -128,12 +155,7 @@
 	                    backing:backingType
 	                      defer:flag];
 
-
-  /* A possible modification here would be to only accept mouse moved events
-   * if any of the child GdkWindows are interested in mouse moved events.
-   */
   [self setAcceptsMouseMovedEvents:YES];
-
   [self setDelegate:self];
   [self setReleasedWhenClosed:YES];
 
@@ -176,13 +198,14 @@
   GdkWindowObject *private = (GdkWindowObject *)window;
   GdkWindowImplQuartz *impl = GDK_WINDOW_IMPL_QUARTZ (private->impl);
 
-  /* FIXME: Is this right? If so, the switch shouldn't be needed. Need
-   * this + some tweaking to the event/grab code to get menus
-   * working...
+  if (!private->accept_focus)
+    return NO;
+
+  /* Popup windows should not be able to get focused in the window
+   * manager sense, it's only handled through grabs.
    */
-  /*if (private->window_type == GDK_WINDOW_TEMP)
+  if (private->window_type == GDK_WINDOW_TEMP)
     return NO;
-  */
 
   switch (impl->type_hint)
     {
